package nu.zoom.dsl.ast;

import nu.zoom.dsl.parser.StatesBaseVisitor;
import nu.zoom.dsl.parser.StatesParser;
import org.antlr.v4.runtime.tree.TerminalNode;

import java.util.*;
import java.util.stream.Stream;

public class StatesVisitorTransformer extends StatesBaseVisitor<StatesParser.DocumentContext> {
    private final HashMap<String,String> config = new HashMap<>();
    private final HashSet<TypeNode> nodeTypes = new HashSet<>();
    private final HashSet<TypeNode> messageTypes = new HashSet<>();
    // from -> <to, message>
    private final HashMap<String,HashMap<String, String>> transitions = new HashMap<>();

    @Override
    public StatesParser.DocumentContext visitTransition(StatesParser.TransitionContext ctx) {
        String from = ctx.from().state().typeName().IDENTIFIER().getText() ;
        String to = ctx.to().state().typeName().IDENTIFIER().getText() ;
        String message = ctx.message().typeName().IDENTIFIER().getText() ;
        this.transitions.computeIfAbsent(from, k -> new HashMap<>()).put(to, message);
        return super.visitTransition(ctx);
    }

    @Override
    public StatesParser.DocumentContext visitState(StatesParser.StateContext ctx) {
        String stateName = ctx.typeName().IDENTIFIER().getText() ;
        List<FieldNode> fields = extractFields(ctx.typeDeclaration()) ;
        this.nodeTypes.add(new TypeNode(stateName, fields));
        return super.visitState(ctx);
    }

    @Override
    public StatesParser.DocumentContext visitMessage(StatesParser.MessageContext ctx) {
        String messageName = ctx.typeName().IDENTIFIER().getText() ;
        List<FieldNode> fields = extractFields(ctx.typeDeclaration()) ;
        this.messageTypes.add(new TypeNode(messageName, fields));
        return super.visitMessage(ctx);
    }

    @Override
    public StatesParser.DocumentContext visitConfigitem(StatesParser.ConfigitemContext ctx) {
        String configKey = ctx.configkey().IDENTIFIER().getText();
        String configValue = getText(ctx.configvalue().IDENTIFIER(), ctx.configvalue().VALUE());
        this.config.put(configKey, configValue);
        return super.visitConfigitem(ctx);
    }

    public Set<StateNode> getStates() {
        HashSet<StateNode> states = new HashSet<>();
        this.transitions.forEach((state,v)->{
            HashSet<TransitionNode> transitionNodes = new HashSet<>();
            v.forEach((to, message) -> transitionNodes.add(new TransitionNode(message, to)));
            states.add(new StateNode(state, "", transitionNodes)) ;
        }) ;
        return states ;
    }

    public Map<String,String> getConfig() {
        return Map.copyOf(config);
    }

    public Set<TypeNode> getTypes() {
        // TODO calculate data types from NodeTypes and MessageTypes with duplicate check.
        HashMap<String, TypeNode> typeNodes = new HashMap<>();
        this.nodeTypes.forEach(typeNode -> {
            if (typeNodes.containsKey(typeNode.name())) {
                throw new RuntimeException("Duplicate type name: " + typeNode.name());
            } else {
                typeNodes.put(typeNode.name(), typeNode);
            }
        }) ;
        this.messageTypes.forEach(typeNode -> {
            if (typeNodes.containsKey(typeNode.name())) {
                throw new RuntimeException("Duplicate type name: " + typeNode.name());
            } else {
                typeNodes.put(typeNode.name(), typeNode);
            }
        }) ;
        return Set.of(typeNodes.values().toArray(new TypeNode[0])) ;
    }

    private List<FieldNode> extractFields(StatesParser.TypeDeclarationContext declaration) {
        if (declaration == null) {
            return Collections.emptyList();
        }
        return declaration
                .typeField()
                .stream()
                .map(
                        ctx ->
                                new FieldNode(ctx.fieldName().getText(), ctx.fieldType().getText())
                )
                .toList();
    }

    // Concatenate the text from two terminal nodes. Useful for contexts that are either an identifier or a value,
    // and you just want the text from whichever is not null.
    private String getText(TerminalNode identifier, TerminalNode value) {
        return
                ((identifier != null) ? identifier.getText() : "") +
                        ((value != null) ? value.getText() : "");
    }

}
